#version 400 core

#define MAXCOUNT 512

#define COLOUR_INTERIOR 0
#define COLOUR_BOUNDARY 1
#define COLOUR_EXTERIOR1 2
#define COLOUR_EXTERIOR2 3
#define COLOUR_MINTERIOR 4
#define COLOUR_MBOUNDARY 5
#define COLOUR_MEXTERIOR1 6
#define COLOUR_MEXTERIOR2 7
#define COLOUR_ACCENT 8
#define COLOURS 9

const double mer = 535.491655524765;

uniform bool centering;
uniform bool mandelbrot;
uniform int count;
uniform dvec2 inflection[MAXCOUNT];
uniform double er[MAXCOUNT];

uniform int iterations;
uniform int samples;

uniform dvec2 center;
uniform dvec2 radius;
uniform dvec2 aspect;

uniform vec4 colours[COLOURS];

in vec2 texcoord;

out vec4 colour;

dvec2 cMul(dvec2 a, dvec2 b)
{
  return dvec2(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

dvec2 cSqr(dvec2 a)
{
  return cMul(a, a);
}

dvec4 cMul(dvec4 a, dvec4 b)
{
  return dvec4(cMul(a.xy, b.xy), cMul(a.xy, b.zw) + cMul(a.zw, b.xy));
}

dvec4 cSqr(dvec4 a)
{
  return cMul(a, a);
}

dvec2 cSqrt(dvec2 z)
{
  double m = length(z);
  double x = sqrt(max(0.0, 0.5 * (m + z.x)));
  double y = sqrt(max(0.0, 0.5 * (m - z.x)));
  return dvec2(x, sign(z.y) * y);
}

uint interleave(uint x, uint y)
{
  uint o = 0u;
  for (int bo = 31; bo >= 0; --bo)
  {
    uint bit = 1u << (bo >> 1);
    o <<= 1;
    o |= (((((bo & 1) == 1) ? x : y) & bit) == bit) ? 1u : 0u;
  }
  return o;
}

// http://www.burtleburtle.net/bob/hash/integer.html
uint hash(uint a)
{
  a = (a+0x7ed55d16u) + (a<<12);
  a = (a^0xc761c23cu) ^ (a>>19);
  a = (a+0x165667b1u) + (a<<5);
  a = (a+0xd3a2646cu) ^ (a<<9);
  a = (a+0xfd7046c5u) + (a<<3);
  a = (a^0xb55a4f09u) ^ (a>>16);
  return a;
}

float uhash(uint a)
{
  return float(hash(a)) / pow(2.0, 32.0);
}

// https://github.com/lycium/FractalTracer/blob/da1168615c3a0dfdb3208d0901952d4a2e0df758/src/renderer/Renderer.h#L108
float tent(float v)
{
  float orig = v * 2.0 - 1.0;
  v = orig / sqrt(abs(orig));
  v = v - sign(orig);
  if (isnan(v) || isinf(v))
  {
    v = 0.0;
  }
  return v;
}

vec2 tent(vec2 x)
{
  return vec2(tent(x.x), tent(x.y));
}

const float g2 = 1.32471795724474602596;
const float a1 = 1.0/g2;
const float a2 = 1.0/(g2*g2);
const uvec2 tile = uvec2(0u, 0u);

float linear_to_srgb(float c)
{
  c = clamp(c, 0.0, 1.0);
  if (c <= 0.0031308)
  {
    return 12.92 * c;
  }
  else
  {
    return 1.055 * pow(c, 1.0 / 2.4) - 0.055;
  }
}

vec3 linear_to_srgb(vec3 x)
{
  return vec3(linear_to_srgb(x.r), linear_to_srgb(x.g), linear_to_srgb(x.b));
}

vec4 linear_to_srgb(vec4 x)
{
  return vec4(linear_to_srgb(x.rgb), x.a);
}

void main()
{
  uint seed = uint(samples) * interleave(tile.x + uint(gl_FragCoord.x), tile.y + uint(gl_FragCoord.y));
  vec2 jseed = vec2(uhash(2u * seed + 0u), uhash(2u * seed + 1u));
  vec2 dx = dFdx(texcoord);
  vec2 dy = dFdy(texcoord);
  double px = double(length(dvec4(radius, radius) * dvec4(dFdx(vec2(aspect) * texcoord), dFdy(vec2(aspect) * texcoord))));
  vec4 colourAccum = vec4(0.0);
  for (int sample_ = 0; sample_ < (samples == 0 ? 1 : samples); ++sample_)
  {
    vec2 jitter = samples == 0 ? vec2(0.0, 0.0) : tent(fract(jseed + vec2(a1, a2) * float(sample_)));
    dvec2 texcoord_ = center + cMul(radius, aspect * dvec2(texcoord + jitter.x * dx + jitter.y * dy));

  dvec4 c = dvec4(texcoord_, px, 0.0);
  vec4 mcolour = vec4(0.0, 0.0, 0.0, 0.0);
  if (mandelbrot)
  {
    int n = 0;
    dvec4 z = c;
    for (n = 0; n < 1024; ++n)
    {
      if (dot(z.xy, z.xy) >= mer)
        break;
      z = cSqr(z) + c;
    }
    if (dot(z.xy, z.xy) < mer)
    {
      mcolour = colours[COLOUR_MINTERIOR];
    }
    else
    {
      float r = float(length(z.xy));
      float dr = float(length(z.zw));
      float de = 2.0 * r * log(r) / dr;
      float g = tanh(clamp(de, 0.0, 4.0));
      if (isnan(de) || isinf(de) || isnan(dr) || isinf(dr) || isnan(r) || isinf(r) || isnan(g) || isinf(g))
        g = 1.0;
      mcolour = mix(colours[COLOUR_MBOUNDARY], z.y > 0 ? colours[COLOUR_MEXTERIOR1] : colours[COLOUR_MEXTERIOR2], g);
    }
  }
  int i;
  bool escaped = false;
  double R = 2.0;
  double R2 = R * R;
  if (count > 0)
  {
    for (i = 0; i < count && i < MAXCOUNT; ++i)
    {
      R2 = er[i];
      if (dot(c.xy, c.xy) >= R2)
      {
        escaped = true;
        break;
      }
      if (centering && i == 0)
      {
        c += dvec4(inflection[0], 0.0, 0.0);
      }
      dvec4 f = dvec4(inflection[i], 0.0, 0.0);
      c = cSqr(c - f) + f;
    }
  }
  int n = 0;
  dvec4 z = count != 0 ? c : dvec4(0.0, 0.0, 0.0, 0.0);
        c = count != 0 ? dvec4(inflection[count - 1], 0.0, 0.0) : c;
  if (! escaped)
  {
    for (n = 0; n < iterations; ++n)
    {
      if (dot(z.xy, z.xy) >= R2)
      {
        escaped = true;
        break;
      }
      z = cSqr(z) + c;
    }
  }
  if (! escaped)
  {
    colourAccum += mix(colours[COLOUR_INTERIOR], vec4(mcolour.rgb, 1.0), mandelbrot ? mcolour.a : 0.0);
    continue;
  }
  float r = float(length(z.xy));
  float dr = float(length(z.zw));
  float de = 2.0 * r * log(r) / dr;
  float g = tanh(clamp(de, 0.0, 4.0));
  if (isnan(de) || isinf(de) || isnan(dr) || isinf(dr) || isnan(r) || isinf(r) || isnan(g) || isinf(g))
    g = 1.0;
  bool decomposition = z.y > 0.0;
  vec4 exterior = decomposition ? colours[COLOUR_EXTERIOR1] : colours[COLOUR_EXTERIOR2];
  float rainbow = pow(2.0, clamp(1.0 - log2(log(float(dot(z.xy, z.xy)))/log(float(R2))), 0.0, 1.0)) - 1.0;
  vec4 erainbow = vec4(mix(mix(vec3(normalize(z.xy), 0.0), vec3(1.0), 0.5), colours[COLOUR_ACCENT].rgb, colours[COLOUR_ACCENT].a), 1.0);
  exterior = mix(exterior, erainbow, rainbow);
  colourAccum += mix(mix(colours[COLOUR_BOUNDARY], exterior, g), vec4(mcolour.rgb, 1.0), mandelbrot ? mcolour.a : 0.0);

  }
  colour = linear_to_srgb(colourAccum / colourAccum.a);
}
