#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fstream>
#include <sstream>
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include "png.h"
#include "settings.h"
#include "shader.h"
#include "state.h"
#include "version.h"

// https://stackoverflow.com/a/2072890
inline bool ends_with(std::string const & value, std::string const & ending)
{
    if (ending.size() > value.size()) return false;
    return std::equal(ending.rbegin(), ending.rend(), value.rbegin());
}

// https://stackoverflow.com/a/31526753
static int mini(int x, int y)
{
    return x < y ? x : y;
}
static int maxi(int x, int y)
{
    return x > y ? x : y;
}
GLFWmonitor* get_current_monitor(GLFWwindow *window)
{
    int nmonitors, i;
    int wx, wy, ww, wh;
    int mx, my, mw, mh;
    int overlap, bestoverlap;
    GLFWmonitor *bestmonitor;
    GLFWmonitor **monitors;
    const GLFWvidmode *mode;

    bestoverlap = 0;
    bestmonitor = NULL;

    glfwGetWindowPos(window, &wx, &wy);
    glfwGetWindowSize(window, &ww, &wh);
    monitors = glfwGetMonitors(&nmonitors);

    for (i = 0; i < nmonitors; i++) {
        mode = glfwGetVideoMode(monitors[i]);
        glfwGetMonitorPos(monitors[i], &mx, &my);
        mw = mode->width;
        mh = mode->height;

        overlap =
            maxi(0, mini(wx + ww, mx + mw) - maxi(wx, mx)) *
            maxi(0, mini(wy + wh, my + mh) - maxi(wy, my));

        if (bestoverlap < overlap) {
            bestoverlap = overlap;
            bestmonitor = monitors[i];
        }
    }

    return bestmonitor;
}

void set_fullscreen(GLFWwindow *window, bool fullscreen)
{
  static int wx = 0, wy = 0, ww = 640, wh = 360; // FIXME
  if (fullscreen)
  {
    glfwGetWindowPos(window, &wx, &wy);
    glfwGetWindowSize(window, &ww, &wh);
    GLFWmonitor *monitor = get_current_monitor(window);
    const GLFWvidmode *mode = glfwGetVideoMode(monitor);
    glfwSetWindowMonitor(window, monitor, 0, 0, mode->width, mode->height, GLFW_DONT_CARE);
  }
  else
  {
    glfwSetWindowMonitor(window, nullptr, wx, wy, ww, wh, GLFW_DONT_CARE);
  }
}

static inline dvec2 cMul(dvec2 a, dvec2 b)
{
  return dvec2(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

std::string make_filename(const char *extension, bool include_frame_number)
{
  char filename[256];
  time_t seconds_since_epoch = S.time;
  struct tm *t = localtime(&seconds_since_epoch);
  if (t)
  {
    // human-readable datetime format
    if (include_frame_number)
    {
      snprintf(filename, sizeof(filename), "%04d-%02d-%02d-%02d-%02d-%02d-%06d%s", 1900 + t->tm_year, t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, S.frame, extension);
    }
    else
    {
      snprintf(filename, sizeof(filename), "%04d-%02d-%02d-%02d-%02d-%02d%s", 1900 + t->tm_year, t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, extension);
    }
  }
  else
  {
    // cryptically-named fallback in case the localtime() function fails
    if (include_frame_number)
    {
      snprintf(filename, sizeof(filename), "%08x-%06d%s", (unsigned int) seconds_since_epoch, S.frame, extension);
    }
    else
    {
      snprintf(filename, sizeof(filename), "%08x%s", (unsigned int) seconds_since_epoch, extension);
    }
  }
  return filename;
}

static void save_screenshot()
{
  // in video mode, save settings only on first frame instead of every frame
  std::string filename = make_filename(".ig.toml", ! S.video);
  if (! S.video || S.frame == 0)
  {
    fprintf(stderr, "saving settings '%s'\n", filename.c_str());
    std::ofstream ofs(filename);
    ofs << S.s;
    ofs.close();
  }
  filename = make_filename(".ig.png", ! S.video);
  if (S.ppm_width * S.ppm_height < S.fbo_width * S.fbo_height)
  {
    if (S.ppm)
    {
      free(S.ppm);
      S.ppm = 0;
    }
    S.ppm_width = S.fbo_width;
    S.ppm_height = S.fbo_height;
  }
  if (! S.ppm)
  {
    S.ppm = (unsigned char *) malloc(S.ppm_width * S.ppm_height * 3);
  }
  if (S.ppm)
  {
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadPixels(0, 0, S.fbo_width, S.fbo_height, GL_RGB, GL_UNSIGNED_BYTE, S.ppm);
    if (S.video)
    {
      // write PPM to stream FIXME TODO check for errors
      fprintf(S.video, "P6\n%d %d\n255\n", S.fbo_width, S.fbo_height);
      for (int y = S.fbo_height - 1; y >= 0; --y)
      {
        fwrite(S.ppm + y * S.fbo_width * 3, S.fbo_width * 3, 1, S.video);
      }
      fflush(S.video);
    }
    if (! S.video || S.frame == 0)
    {
      // write PNG file
      fprintf(stderr, "saving image    '%s'\n", filename.c_str());
      std::ostringstream comment;
      comment << S.s;
      if (0 != save_png(filename.c_str(), S.ppm, S.fbo_width, S.fbo_height, comment.str().c_str()))
      {
        fprintf(stderr, "saving image   '%s' FAILED\n", filename.c_str());
      }
    }
  }
  else
  {
    fprintf(stderr, "saving image FAILED\n");
  }
  S.frame++;
}

static void fbresizecb(GLFWwindow *window, int width, int height)
{
  (void) window;
  S.fb_width = width;
  S.fb_height = height;
  glViewport(0, 0, width, height);
}

static void resizecb(GLFWwindow *window, int width, int height)
{
  (void) window;
  S.s.width = width;
  S.s.height = height;
  S.aspect = dvec2(S.s.width * 1.0 / S.s.height, 1.0);
}

static void maximizecb(GLFWwindow *window, int maximized)
{
  int width, height;
  glfwGetWindowSize(window, &width, &height);
  resizecb(window, width, height);
  S.s.fullscreen = maximized;
}

static dvec2 mouse_position(double x, double y)
{
  double w = S.s.width;
  double h = S.s.height;
  double dx = 2 * ((x + 0.5) / w - 0.5);
  double dy = 2 * (0.5 - (y + 0.5) / h);
  dvec2 p = dvec2(dx, dy);
  dvec2 t = S.s.center + cMul(dvec2(S.s.radius, 0.0), S.aspect * p);
  if (S.s.centering && S.s.inflection > 0)
    t += S.s.inflections[S.s.inflection - 1];
  return t;
}

static void zoom(double x, double y, double amount)
{
  double g = pow(0.5, amount / 5);
  dvec2 center = S.s.center;
  if (S.s.centering && S.s.inflection > 0)
    center += S.s.inflections[S.s.inflection - 1];
  center = mix(mouse_position(x, y), center, g);
  if (S.s.centering && S.s.inflection > 0)
    center -= S.s.inflections[S.s.inflection - 1];
  S.s.center = center;
  S.s.radius *= g;
}

static void scrollcb(GLFWwindow* window, double xoffset, double yoffset)
{
  (void) xoffset;
  double x = 0, y = 0;
  glfwGetCursorPos(window, &x, &y);
  zoom(x, y, yoffset);
}

static void start_playback()
{
  int keyframes;
  if (S.is_stop_motion)
  {
    S.stop_motion_inflection = S.s.inflection;
    keyframes = S.s.inflections.size() + 1;
  }
  else
  {
    keyframes = S.s.keyframes.size();
  }
  S.aframes = round(keyframes / S.s.speed);
  S.s.speed = double(keyframes) / double(S.aframes);
  if (isnan(S.s.speed) || isinf(S.s.speed))
  {
    S.s.speed = 1.0 / 64;
  }
  S.aindex = -S.s.speed;
  S.playing = true;
}

static void keycb(GLFWwindow *window, int key, int scancode, int action, int mods)
{
  (void) scancode;
  if (action == GLFW_PRESS)
    switch (key)
    {
      case GLFW_KEY_Q:
      case GLFW_KEY_ESCAPE:
        glfwSetWindowShouldClose(window, GL_TRUE);
        break;
      case GLFW_KEY_F11:
        S.s.fullscreen = ! S.s.fullscreen;
        set_fullscreen(window, S.s.fullscreen);
        break;
      // keyboard navigation
      case GLFW_KEY_HOME:
        S.s.center = dvec2(0.0, 0.0);
        S.s.radius = 2.0;
        break;
      case GLFW_KEY_UP:
        scrollcb(window, 0, (mods & GLFW_MOD_SHIFT) ? 10 : 1);
        break;
      case GLFW_KEY_DOWN:
        scrollcb(window, 0, (mods & GLFW_MOD_SHIFT) ? -10 : -1);
        break;
      case GLFW_KEY_PAGE_UP:   zoom(S.s.width * 0.5, S.s.height * 0.5, (mods & GLFW_MOD_SHIFT) ? 5 : 1);  break;
      case GLFW_KEY_PAGE_DOWN: zoom(S.s.width * 0.5, S.s.height * 0.5, (mods & GLFW_MOD_SHIFT) ? -5 : -1);  break;
      case GLFW_KEY_1: case GLFW_KEY_KP_1: zoom(0, S.s.height, 5); break;
      case GLFW_KEY_2: case GLFW_KEY_KP_2: zoom(S.s.width * 0.5, S.s.height, 5); break;
      case GLFW_KEY_3: case GLFW_KEY_KP_3: zoom(S.s.width, S.s.height, 5); break;
      case GLFW_KEY_4: case GLFW_KEY_KP_4: zoom(0, S.s.height * 0.5, 5); break;
      case GLFW_KEY_5: case GLFW_KEY_KP_5: zoom(S.s.width * 0.5, S.s.height * 0.5, 5); break;
      case GLFW_KEY_6: case GLFW_KEY_KP_6: zoom(S.s.width, S.s.height * 0.5, 5); break;
      case GLFW_KEY_7: case GLFW_KEY_KP_7: zoom(0, 0, 5); break;
      case GLFW_KEY_8: case GLFW_KEY_KP_8: zoom(S.s.width * 0.5, 0, 5); break;
      case GLFW_KEY_9: case GLFW_KEY_KP_9: zoom(S.s.width, 0, 5); break;
      case GLFW_KEY_0: case GLFW_KEY_KP_0: zoom(S.s.width * 0.5, S.s.height * 0.5, -5); break;
      // other keyboard controls
      case GLFW_KEY_M:
        S.s.mandelbrot = ! S.s.mandelbrot;
        break;
      case GLFW_KEY_C:
        S.s.centering = ! S.s.centering;
        break;
      case GLFW_KEY_MINUS:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.s.inflection = 0;
        }
        else
        {
          undo_add_point();
        }
        break;
      case GLFW_KEY_EQUAL:
        redo_add_point();
        break;
      case GLFW_KEY_D:
        if (S.have_double)
          S.s.use_double = true;
        else
          fprintf(stderr, "double precision not available (needs OpenGL 4)\n");
        break;
      case GLFW_KEY_F:
        S.s.use_double = false;
        break;
      case GLFW_KEY_J:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.s.samples = 0;
        }
        else
        {
          if (S.s.samples < 256) // FIXME
          {
            S.s.samples <<= 1;
            S.s.samples += ! S.s.samples;
          }
        }
        break;
      case GLFW_KEY_U:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.s.undersample = 1;
        }
        else
        {
          S.s.undersample += 1;
        }
        break;
      case GLFW_KEY_O:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.s.oversample = 1;
        }
        else
        {
          S.s.oversample += 1;
        }
        break;
      case GLFW_KEY_I:
        S.s.undersample = 1;
        S.s.oversample = 1;
        break;
      case GLFW_KEY_N:
        if (mods & GLFW_MOD_SHIFT)
        {
          if (S.s.iterations > (1 << 4)) // FIXME
          {
            S.s.iterations >>= 1;
          }
        }
        else
        {
          if (S.s.iterations < (1 << 24)) // FIXME
          {
            S.s.iterations <<= 1;
          }
        }
        break;
      case GLFW_KEY_S:
        // save after rendering
        S.screenshot = true;
        break;
      case GLFW_KEY_T:
        S.s.theme = (S.s.theme + 1) % S.s.themes.size();;
        break;
      case GLFW_KEY_A:
        if (mods & GLFW_MOD_SHIFT)
        {
          if (S.s.themes.size() > 0)
          S.s.themes[0].colours[COLOUR_ACCENT] = lighttheme.colours[COLOUR_ACCENT];
          if (S.s.themes.size() > 1)
          S.s.themes[1].colours[COLOUR_ACCENT] = darktheme.colours[COLOUR_ACCENT];
        }
        else
        {
          for (int c = 0; c < 4; ++c)
          {
            double x = 0;
            if (S.s.themes.size() > 0)
            S.s.themes[0].colours[COLOUR_ACCENT][c] = x = rand() / (double) RAND_MAX;
            if (S.s.themes.size() > 1)
            S.s.themes[1].colours[COLOUR_ACCENT][c] = x;
          }
        }
        break;
      case GLFW_KEY_B:
        if (mods & GLFW_MOD_SHIFT)
        {
          if (S.s.themes.size() > 0)
          {
          S.s.themes[0].colours[COLOUR_EXTERIOR1] = lighttheme.colours[COLOUR_EXTERIOR1];
          S.s.themes[0].colours[COLOUR_EXTERIOR2] = lighttheme.colours[COLOUR_EXTERIOR2];
          }
          if (S.s.themes.size() > 1)
          {
          S.s.themes[1].colours[COLOUR_EXTERIOR1] = darktheme.colours[COLOUR_EXTERIOR1];
          S.s.themes[1].colours[COLOUR_EXTERIOR2] = darktheme.colours[COLOUR_EXTERIOR2];
          }
        }
        else
        {
          for (int c = 0; c < 3; ++c)
          {
            double x = 0;
            if (S.s.themes.size() > 0)
            S.s.themes[0].colours[COLOUR_EXTERIOR1][c] = x = rand() / (double) RAND_MAX;
            if (S.s.themes.size() > 1)
            S.s.themes[1].colours[COLOUR_EXTERIOR1][c] = x;
            if (S.s.themes.size() > 0)
            S.s.themes[0].colours[COLOUR_EXTERIOR2][c] = x = rand() / (double) RAND_MAX;
            if (S.s.themes.size() > 1)
            S.s.themes[1].colours[COLOUR_EXTERIOR2][c] = x;
          }
        }
        break;
      case GLFW_KEY_SPACE:
        add_keyframe();
        break;
      case GLFW_KEY_DELETE:
        delete_keyframe();
        break;
      case GLFW_KEY_LEFT:
        previous_keyframe();
        break;
      case GLFW_KEY_RIGHT:
        next_keyframe();
        break;
      case GLFW_KEY_ENTER:
      case GLFW_KEY_KP_ENTER:
        S.playing = ! S.playing;
        if (S.playing)
        {
          start_playback();
        }
        else
        {
          if (S.is_stop_motion)
          {
            S.s.inflection = S.stop_motion_inflection;
          }
          else
          {
            activate_keyframe();
          }
        }
        break;
      case GLFW_KEY_RIGHT_BRACKET:
        faster_animation(mods & GLFW_MOD_SHIFT);
        break;
      case GLFW_KEY_LEFT_BRACKET:
        slower_animation(mods & GLFW_MOD_SHIFT);
        break;
      case GLFW_KEY_P:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.s.speed = 1.0;
        }
        else
        {
          S.s.speed = 1.0 / 64;
        }
        break;
      case GLFW_KEY_V:
        if (mods & GLFW_MOD_SHIFT)
        {
          S.time = time(0);
          S.frame = 0;
          start_playback();
          S.recording = true;
          const char *video = getenv("IG_VIDEO");
          if (video)
          {
            std::string filename = make_filename(".ig", false);
            size_t bytes = strlen(video) + strlen(filename.c_str()) + 256;
            char *command = (char *) malloc(bytes);
            if (command)
            {
              snprintf(command, bytes, video, filename.c_str());
              S.video = popen(command, "w");
              if (S.video)
              {
                fprintf(stderr, "saving video '%s'\n", command);
              }
              else
              {
                fprintf(stderr, "saving video '%s' FAILED\n", command);
              }
              free(command);
            }
            else
            {
              fprintf(stderr, "saving video FAILED\n");
            }
          }
        }
        else
        {
          if (! S.playing)
          {
            S.is_stop_motion = ! S.is_stop_motion;
          }
        }
        break;
      case GLFW_KEY_H:
        fprintf(stderr,
          "keyboard controls for navigation:\n"
          "\tESC, Q\tquit\n"
          "\tH\tshow this help\n"
          "\tF11\ttoggle fullscreen\n"
          "\tHOME\treset view\n"
          "\tUP\tzoom in about cursor (faster with SHIFT)\n"
          "\tDOWN\tzoom out about cursor (faster with SHIFT)\n"
          "\tPGUP\tzoom in about center (faster with SHIFT)\n"
          "\tPGDOWN\tzoom out about center (faster with SHIFT)\n"
          "\t1-9\tzoom in to quadrant\n"
          "\t0\tzoom out\n"
          "\tM\ttoggle Mandelbrot overlay\n"
          "\tC\ttoggle centering mode\n"
          "\tN\tincrease iteration count (decrease with SHIFT)\n"
          "\tJ\tincrease sample count (reset with SHIFT)\n"
          "\tU\tincrease image undersampling (reset with SHIFT)\n"
          "\tO\tincrease image oversampling (reset with SHIFT)\n"
          "\tI\treset image sampling ratio\n"
          "\tA\trandomize accent colour (reset with SHIFT)\n"
          "\tB\trandomize decomposition colours (reset with SHIFT)\n"
          "\tD\tuse double precision (requires OpenGL 4)\n"
          "\tF\tuse single precision (default)\n"
          "\t-\tundo add inflection point (remove all with SHIFT)\n"
          "\t=\tredo add inflection point\n"
          "\tT\tchange colour theme\n"
          "\tS\tsave screenshot (in PNG format)\n"
          "keyboard controls for animation:\n"
          "\tSPACE\tadd keyframe\n"
          "\tDELETE\tdelete keyframe\n"
          "\tLEFT\tactivate previous keyframe\n"
          "\tRIGHT\tactivate next keyframe\n"
          "\tENTER\tplay animation (press again to stop)\n"
          "\t[\tplay slower (more with SHIFT)\n"
          "\t]\tplay faster (more with SHIFT)\n"
          "\tP\treset speed to default (set to 1.0 with SHIFT)\n"
          "\tV\t(without SHIFT) toggle video mode (stopframe, keyframe)\n"
          "\tV\t(with SHIFT) save video sequence\n"
          "mouse controls:\n"
          "\tWHEEL\tzoom about mouse cursor position\n"
          "\tLEFT\tadd inflection point at cursor position\n"
          "\tRIGHT\tundo add inflection point\n"
          "\tMIDDLE\trecenter window about mouse cursor position\n"
        );
        break;
    }
}

static void buttoncb(GLFWwindow *window, int button, int action, int mods)
{
  (void) mods;
  if (action == GLFW_RELEASE)
  {
    double x = 0, y = 0;
    glfwGetCursorPos(window, &x, &y);
    switch (button)
    {
      case GLFW_MOUSE_BUTTON_LEFT:
        add_point(mouse_position(x, y));
        break;
      case GLFW_MOUSE_BUTTON_RIGHT:
        undo_add_point();
        break;
      case GLFW_MOUSE_BUTTON_MIDDLE:
        dvec2 center = mouse_position(x, y);
        if (S.s.centering && S.s.inflection > 0)
          center -= S.s.inflections[S.s.inflection - 1];
        S.s.center = center;
        break;
    }
  }
}

void draw()
{
  if (S.s.undersample < 1) S.s.undersample = 1;
  int image_width = S.fb_width * S.s.oversample / S.s.undersample;
  int image_height = S.fb_height * S.s.oversample / S.s.undersample;
  if (image_width > 1 << 13) image_width = 1 << 13;
  if (image_height > 1 << 13) image_height = 1 << 13;
  if (image_width < 1 << 4) image_width = 1 << 4;
  if (image_height < 1 << 4) image_height = 1 << 4;
  if (image_width != S.fbo_width || image_height != S.fbo_height)
  {
    // recreate framebuffer
    S.fbo_width = image_width;
    S.fbo_height = image_height;
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, S.fbo_width, S.fbo_height, 0, GL_RGB, GL_UNSIGNED_BYTE, nullptr);
    glBindFramebuffer(GL_FRAMEBUFFER, S.fbo);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, S.texture, 0);
  }
  glBindFramebuffer(GL_FRAMEBUFFER, S.fbo);
  glViewport(0, 0, S.fbo_width, S.fbo_height);
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
  glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
  glBlitFramebuffer(0, 0, S.fbo_width, S.fbo_height, 0, 0, S.fb_width, S.fb_height, GL_COLOR_BUFFER_BIT, GL_NEAREST);
}

extern int main(int argc, char **argv)
{
  fprintf(stderr, "inflector-gadget %s (GPL) 2023 Claude Heiland-Allen\n", inflector_gadget_version_string);

  S.s = default_settings;
  for (int arg = 1; arg < argc; ++arg)
  {
    if (ends_with(argv[arg], ".ig.toml"))
    {
      std::ifstream ifs(argv[arg]);
      ifs >> S.s;
      ifs.close();
    }
    else if (ends_with(argv[arg], ".png"))
    {
      std::string comment = read_png_comment(argv[arg]);
      if (comment != "")
      {
        std::istringstream iss(comment);
        iss >> S.s;
      }
    }
  }

  S.ppm = nullptr;
  S.ppm_width = 0;
  S.ppm_height = 0;
  S.max_count = 0;
  S.time = time(0);
  srand(S.time);
  S.frame = 0;
  S.aindex = 0;
  S.playing = false;
  S.fbo_width = 0;
  S.fbo_height = 0;

  glfwInit();
  glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
  glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
  glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
  glfwWindowHint(GLFW_RESIZABLE, GL_TRUE);
  GLFWwindow *window = glfwCreateWindow(S.s.width, S.s.height, "inflector-gadget", 0, 0);
  if (! window)
  {
    fprintf(stderr, "could not initialize OpenGL 4.0\n");
    glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
    glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
    window = glfwCreateWindow(S.s.width, S.s.height, "inflector-gadget", 0, 0);
    if (! window)
    {
      fprintf(stderr, "could not initialize OpenGL 3.3, giving up\n");
      return 1;
    }
    else
    {
      fprintf(stderr, "initialized OpenGL 3.3\n");
      S.have_double = false;
    }
  }
  else
  {
    fprintf(stderr, "initialized OpenGL 4.0\n");
    S.have_double = true;
  }
  glfwMakeContextCurrent(window);
  glewExperimental = GL_TRUE;
  glewInit();
  glGetError(); // discard common error from glew
  glfwSwapInterval(1);
  glfwSetWindowMaximizeCallback(window, maximizecb);
  glfwSetWindowSizeCallback(window, resizecb);
  int w, h;
  glfwGetFramebufferSize(window, &w, &h);
  fbresizecb(window, w, h);
  glfwSetFramebufferSizeCallback(window, fbresizecb);
  glfwSetMouseButtonCallback(window, buttoncb);
  glfwSetScrollCallback(window, scrollcb);
  glfwSetKeyCallback(window, keycb);
  if (S.s.fullscreen)
  {
    set_fullscreen(window, S.s.fullscreen);
  }
  else
  {
    maximizecb(window, 0);
  }
  S.aspect = dvec2(S.s.width * 1.0 / S.s.height, 1.0);

  S.f = compile_shader(false);
  if (S.have_double)
  {
    S.d = compile_shader(true);
  }

  GLuint vao = 0;
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);

  float vdata[8] =
    { -1.0, -1.0
    ,  1.0, -1.0
    , -1.0,  1.0
    ,  1.0,  1.0
    };
  GLuint vbo = 0;
  glGenBuffers(1, &vbo);
  glBindBuffer(GL_ARRAY_BUFFER, vbo);
  glBufferData(GL_ARRAY_BUFFER, 8 * sizeof(float), &vdata[0], GL_STATIC_DRAW);
  glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray(0);

  glGenTextures(1, &S.texture);
  glBindTexture(GL_TEXTURE_2D, S.texture);
  glGenFramebuffers(1, &S.fbo);

  while (1) {
    if (S.playing)
    {
      update_animation();
      if (S.s.inflection > S.max_count) S.max_count = S.s.inflection;
      update_shader(S.have_double && S.s.use_double ? S.d : S.f, S.s, S.aspect);
      draw();
      if (S.recording)
      {
        save_screenshot();
        if (S.frame >= S.aframes)
        {
          if (S.video)
          {
            pclose(S.video);
            S.video = 0;
          }
          S.recording = false;
          S.playing = false;
          S.time = time(0);
          S.frame = 0;
          if (S.is_stop_motion)
          {
            S.s.inflection = S.stop_motion_inflection;
          }
          else
          {
            activate_keyframe();
          }
        }
      }
      glfwSwapBuffers(window);
      glfwPollEvents();
    }
    else
    {
      if (S.s.inflection > S.max_count) S.max_count = S.s.inflection;
      update_shader(S.have_double && S.s.use_double ? S.d : S.f, S.s, S.aspect);
      draw();
      if (S.screenshot)
      {
        save_screenshot();
        S.screenshot = false;
      }
      glfwSwapBuffers(window);
      glfwWaitEvents();
    }
    if (glfwWindowShouldClose(window))
      break;
  }
  free(S.ppm);

  glfwDestroyWindow(window);
  glfwTerminate();
  return 0;
}
