#include "settings.h"
#include "shader.h"

#include "float_vert.glsl.c"
#include "float_frag.glsl.c"
#include "double_vert.glsl.c"
#include "double_frag.glsl.c"

static void debug_program(GLuint program) {
  GLint status = 0;
  glGetProgramiv(program, GL_LINK_STATUS, &status);
  GLint length = 0;
  glGetProgramiv(program, GL_INFO_LOG_LENGTH, &length);
  char *info = 0;
  if (length) {
    info = (char *) malloc(length + 1);
    info[0] = 0;
    glGetProgramInfoLog(program, length, 0, info);
    info[length] = 0;
  }
  if ((info && info[0]) || ! status) {
    fprintf(stderr, "program link info:\n%s", info ? info : "(no info log)");
  }
  if (info) {
    free(info);
  }
}

static void debug_shader(GLuint shader, GLenum type, const char *source) {
  GLint status = 0;
  glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
  GLint length = 0;
  glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &length);
  char *info = 0;
  if (length) {
    info = (char *) malloc(length + 1);
    info[0] = 0;
    glGetShaderInfoLog(shader, length, 0, info);
    info[length] = 0;
  }
  if ((info && info[0]) || ! status) {
    const char *type_str = "unknown";
    switch (type) {
      case GL_VERTEX_SHADER: type_str = "vertex"; break;
      case GL_FRAGMENT_SHADER: type_str = "fragment"; break;
      case GL_COMPUTE_SHADER: type_str = "compute"; break;
    }
    fprintf(stderr, "%s shader compile info:\n%s\nshader source:\n%s", type_str, info ? info : "(no info log)", source ? source : "(no source)");
  }
  if (info) {
    free(info);
  }
}

static GLuint vertex_fragment_shader(const char *vert, const char *frag) {
  GLuint program = glCreateProgram();
  {
    GLuint shader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(shader, 1, &vert, 0);
    glCompileShader(shader);
    debug_shader(shader, GL_VERTEX_SHADER, vert);
    glAttachShader(program, shader);
    glDeleteShader(shader);
  }
  {
    GLuint shader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(shader, 1, &frag, 0);
    glCompileShader(shader);
    debug_shader(shader, GL_FRAGMENT_SHADER, frag);
    glAttachShader(program, shader);
    glDeleteShader(shader);
  }
  glLinkProgram(program);
  debug_program(program);
  return program;
}

void update_shader(const struct shader &d, const struct settings &s, dvec2 aspect)
{
  if (d.double_precision)
  {
    glUseProgram(d.program);
    glUniform1i (d.centering, s.centering);
    glUniform2dv(d.center, 1, &s.center[0]);
    dvec2 radius = dvec2(s.radius, 0.0);
    glUniform2dv(d.radius, 1, &radius[0]);
    glUniform2dv(d.aspect, 1, &aspect[0]);
    glUniform1i (d.mandelbrot, s.mandelbrot);
    glUniform1i (d.count, s.inflection);
    if (s.inflection > 0)
    {
      dvec2 inflections[MAXCOUNT];
      for (int i = 0; i < s.inflection; ++i)
      {
        inflections[MAXCOUNT - 1 - i] = s.inflections[i];
      }
      glUniform2dv(d.inflection, s.inflection, &inflections[MAXCOUNT - s.inflection][0]);
    }
    glUniform1i (d.iterations, s.iterations);
    glUniform1i (d.samples, s.samples);
    glUniform4fv(d.colours, COLOURS, &s.themes[s.theme].colours[0][0]);
    /* calculate escape radii */
    double er[MAXCOUNT];
    double er0 = 2.0;
    if (s.inflection > 0)
    {
      double c = length(s.inflections[0]);
      double e = (1 + 2 * c) / 2;
      er0 = sqrt(e * e - c * c + c) + e;
      er0 = fmax(er0, 2.0);
      er[MAXCOUNT - 1] = er0;
      for (int i = 1; i < s.inflection; ++i)
      {
        double c = length(s.inflections[i]);
        er[MAXCOUNT - i - 1] = fmax(2.0, sqrt(er[MAXCOUNT - i] + c) + c);
      }
      for (int i = 0; i < s.inflection; ++i)
      {
        er[MAXCOUNT - i - 1] *= er[MAXCOUNT - i - 1];
      }
      glUniform1dv(d.er, s.inflection, &er[MAXCOUNT - s.inflection]);
    }
  }
  else
  {
    vec2 center = s.center;
    vec2 radius = vec2(s.radius, 0.0);
    vec2 faspect = aspect;
    glUseProgram(d.program);
    glUniform1i (d.centering, s.centering);
    glUniform2fv(d.center, 1, &center[0]);
    glUniform2fv(d.radius, 1, &radius[0]);
    glUniform2fv(d.aspect, 1, &faspect[0]);
    glUniform1i (d.mandelbrot, s.mandelbrot);
    glUniform1i (d.count, s.inflection);
    if (s.inflection > 0)
    {
      vec2 inflections[MAXCOUNT];
      for (int i = 0; i < s.inflection; ++i)
      {
        inflections[MAXCOUNT - 1 - i] = s.inflections[i];
      }
      glUniform2fv(d.inflection, s.inflection, &inflections[MAXCOUNT - s.inflection][0]);
    }
    glUniform1i (d.iterations, s.iterations);
    glUniform1i (d.samples, s.samples);
    glUniform4fv(d.colours, COLOURS, &s.themes[s.theme].colours[0][0]);
    /* calculate escape radii */
    float er[MAXCOUNT];
    double er0 = 2.0;
    if (s.inflection > 0)
    {
      double c = length(s.inflections[0]);
      double e = (1 + 2 * c) / 2;
      er0 = sqrt(e * e - c * c + c) + e;
      er0 = fmaxf(er0, 2.0);
      er[MAXCOUNT - 1] = er0;
      for (int i = 1; i < s.inflection; ++i)
      {
        double c = length(s.inflections[i]);
        er[MAXCOUNT - i - 1] = er0 = fmax(2.0, sqrt(er0 + c) + c);
      }
      for (int i = 0; i < s.inflection; ++i)
      {
        er[MAXCOUNT - i - 1] *= er[MAXCOUNT - i - 1];
      }
      glUniform1fv(d.er, s.inflection, &er[MAXCOUNT - s.inflection]);
    }
  }
}

struct shader compile_shader(bool double_precision)
{
  struct shader d;
  d.double_precision = double_precision;
  if (double_precision)
  {
    d.program = vertex_fragment_shader(double_vert, double_frag);
  }
  else
  {
    d.program = vertex_fragment_shader(float_vert, float_frag);
  }
  d.centering  = glGetUniformLocation(d.program, "centering"); 
  d.center     = glGetUniformLocation(d.program, "center"); 
  d.radius     = glGetUniformLocation(d.program, "radius"); 
  d.aspect     = glGetUniformLocation(d.program, "aspect"); 
  d.mandelbrot = glGetUniformLocation(d.program, "mandelbrot"); 
  d.count      = glGetUniformLocation(d.program, "count"); 
  d.inflection = glGetUniformLocation(d.program, "inflection"); 
  d.er         = glGetUniformLocation(d.program, "er"); 
  d.iterations = glGetUniformLocation(d.program, "iterations"); 
  d.samples    = glGetUniformLocation(d.program, "samples"); 
  d.colours    = glGetUniformLocation(d.program, "colours");
  return d;
}
